{-# LANGUAGE TypeSynonymInstances #-}

module DrawStuff (
  display,
  Command (..),
  Pen (..), black, white,
  Distance, Angle
) where

import Test.QuickCheck
    ( elements, oneof, sized, Arbitrary(arbitrary) )

import qualified Data.ByteString.Lazy as B
import Data.Binary.Put

infixr 5 :#:

-- Points

data Pnt = Pnt Float Float
  deriving (Eq,Ord,Show)

instance Num Pnt where
  Pnt x y + Pnt x' y'  =  Pnt (x+x') (y+y')
  Pnt x y - Pnt x' y'  =  Pnt (x-x') (y-y')
  Pnt x y * Pnt x' y'  =  Pnt (x*x') (y*y')
  fromInteger          =  scalar . fromInteger
  abs (Pnt x y)        =  Pnt (abs x) (abs y)
  signum (Pnt x y)     =  Pnt (signum x) (signum y)

instance Fractional Pnt where
  Pnt x y / Pnt x' y'  =  Pnt (x/x') (y/y')
  fromRational         =  scalar . fromRational

scalar :: Float -> Pnt
scalar x  =  Pnt x x

scalarMax :: Pnt -> Pnt
scalarMax (Pnt x y)  =  scalar (x `max` y)

lub :: Pnt -> Pnt -> Pnt
Pnt x y `lub` Pnt x' y'  =  Pnt (x `max` x') (y `max` y')

glb :: Pnt -> Pnt -> Pnt
Pnt x y `glb` Pnt x' y'  =  Pnt (x `min` x') (y `min` y')


-- Colors

data Pen = Colour Float Float Float 
         | Inkless
           deriving (Eq, Ord, Show)

white, black :: Pen
white = Colour 256 256 256
black = Colour 0.0 0.0 0.0

-- Lines

data Ln = Ln Pen Pnt Pnt
  deriving (Eq,Ord,Show)


-- Commands for moving the turtle around
--  Turtles turn counter-clockwise and start facing up

type Angle    = Float
type Distance = Float
type Turtle   = (Pen,Angle,Pnt)

data Command = Go Distance
             | Turn Angle 
             | Sit
             | Command :#: Command
             | Branch Command
             | GrabPen Pen
               deriving (Eq, Ord, Show)

execute :: Command -> [Ln]
execute c  =  lines
  where
  (lines, _)  =  f c (black, 0, Pnt 0 0)

  f :: Command -> Turtle -> ([Ln], Turtle)
  f (c :#: d) turtle             =  (clines ++ dlines, dturtle)
                                    where
                                    (clines, cturtle) = f c turtle
                                    (dlines, dturtle) = f d cturtle
  f (Branch c) turtle            =  (clines, turtle)
                                    where
                                    (clines, _) = f c turtle
  f (Go dst) (pen,ang,pnt)       =  ( [Ln pen pnt endpnt | pen /= Inkless], (pen,ang,endpnt))
                                    where
                                    endpnt = pnt + scalar dst * polar ang
  f (Turn delta) (pen,ang,pnt)   =  ([], (pen,ang-delta,pnt))
  f (GrabPen new) (_,ang,pnt)  =  ([], (new,ang,pnt))
  f Sit turtle                 =  ([], turtle)


-- Rescales all points in a list of lines
--  from an arbitrary scale
--  to (-1.-1) - (1.1)

rescale :: [Ln] -> [Ln]
rescale lines | null points  = []
              | otherwise    = map f lines
  where
    f (Ln pen p q)  =  Ln pen (g p) (g q)
    g p             =  swap ((p - p0) / s)
    points          =  [ r | Ln _ p q <- lines, r <- [p, q] ]
    hi              =  foldr1 lub points
    lo              =  foldr1 glb points
    s               =  scalarMax (hi - lo) * scalar 0.55
    p0              =  (hi + lo) * scalar 0.5
    swap (Pnt x y)  =  Pnt y x

polar :: Angle -> Pnt
polar ang  =  Pnt (cos radians) (sin radians)
  where
  radians  =  ang * 2 * pi / 360

--writing file

parseLine :: Ln -> String
parseLine (Ln (Colour r g b) (Pnt x1 y1) (Pnt x2 y2))
  = 
    "\n<line x1=\"" ++ x1' ++ "%\"" ++
          "y1=\"" ++ y1' ++ "%\"" ++
          "x2=\"" ++ x2' ++ "%\"" ++
          "y2=\"" ++ y2' ++ "%\"" ++
          "style=\"stroke-width:50;shape-rendering:crispEdges;stroke:rgb" ++ rgb ++ "\" />"
  where
    x1' = show $ scaleToSvg x1
    y1' = show $ scaleToSvg (-y1)
    x2' = show $ scaleToSvg x2
    y2' = show $ scaleToSvg (-y2) 
    rgb = "(" ++ show (r) ++ "," ++ show (g) ++"," ++ show (b) ++ ")"

scaleToSvg x =  (x + 1) / 2 * 100.0
    
toHtmlBody :: Command -> String
toHtmlBody = concatMap parseLine . rescale . execute

wrapSvgTag:: Int -> Int -> String -> String
wrapSvgTag n num x | n /= ((n+1) `mod` num) = "<meta http-equiv=\"refresh\" content=\"0.05;url=output" ++ (show ((n+1) `mod` num)) ++ ".html\" \n /><svg height=\"" ++ "600" ++ "\" width=\"" ++ "600" ++ "\">" ++ x ++ "</svg>"
                   | otherwise = "<svg height=\"" ++ "600" ++ "\" width=\"" ++ "600" ++ "\">" ++ x ++ "</svg>"

display :: (Int,Command) -> Int -> IO ()
display (n,com) num = (writeFile ("output/output" ++ (show n) ++ ".html") . (wrapSvgTag n num)) (toHtmlBody (com))