--Voronoi Diagrams

module Main ( main
) where

import Data.List
import Data.Maybe
import DrawStuff

type PointList = [(Int,Int,Int,Int,Pen)]

--main function
main :: IO()
main = do animation frames frames

animation :: Int -> Int -> IO()
animation 1 num = do 
  (display (fullPic 0) num)
animation n num = do 
  (display (fullPic (n-1)) num)
  animation (n-1) num

fullPic :: Int -> (Int,Command)
fullPic n | tiled = (n,(tileImage (pointsOnFrame n)))
          | otherwise = (n,(drawImage (pointsOnFrame n)))

--constant values

noPoints = 3
dims = 20
fadeOff = 4
frames = 40
tiled = True
startingSeed = 82

--some example seeds named based on what they look like:
--72 - neapolitan
--73 - neon
--75 - toxic
--80 - jam toast
--82 - interweaving


--joins commands together
join :: [Command] -> Command
join =  foldr (:#:) Sit

--"randomly" picks some points
startingPoints :: Int -> PointList
startingPoints seed = [ ((random dims (seed+3*s)),(random dims (seed+7*s+3)),((random 4 (seed+9*s+23))-2),((random 4 (seed+13*s))-2),(Colour (random 256 (seed+17*s+79)) (random 256 (seed+29*s+43789)) (random 256 (seed+123*s+83729)))) | s <- [0..(noPoints-1)] ] 
  where
  random :: Num a => Int -> Int -> a
  random max seed = fromIntegral ((9466206131*(((seed)^3)*7389919517)+6214686703) `mod` max)

--generates a list of moving points and updates them each frame
updatePointList :: PointList -> PointList
updatePointList []  = [ (x,y,vx,vy,col) | (x,y,vx,vy,col) <- (startingPoints startingSeed)]
updatePointList (p:ps) = [ (((x+vx) `mod` dims),((y+vy) `mod` dims),vx,vy,col) | (x,y,vx,vy,col) <- (p:ps) ]

--finds where the points are on a given frame by updating the starting points until the desired frame is reached
pointsOnFrame :: Int -> PointList
pointsOnFrame 0 = updatePointList []
pointsOnFrame n = (foldr1 (.) (replicate (n+1) updatePointList)) []

--turns the pointlist into a turtle command
drawImage :: PointList -> Command
drawImage [] = Sit
drawImage pl = (join [ Branch (join [ (GrabPen (closestColour pl (x,y) ) :#: Go 1) | y <- [0 ..(dims-1)] ] ) :#: GrabPen Inkless :#: Turn 90 :#: Go 1 :#: Turn (-90) | x <- [0..(dims-1)] ]) :#: GrabPen white :#: Go (fromIntegral(dims + 1))

--when tiled is True, the generated images will be tiled for a clearer demonstration of the colour wraparound.
tileImage com = copy 2 (Branch (copy 2 (Branch (drawImage com) :#: GrabPen Inkless :#: Go (fromIntegral dims))) :#: Turn 90 :#: GrabPen Inkless :#: Go (fromIntegral dims) :#: Turn (-90))
  where
  copy x com = foldr (:#:) Sit (replicate x com)

--calculates what colour each pixel should be by finding the closest point
closestColour :: PointList -> (Int,Int) -> Pen
closestColour pl pt = minDist  (distances pt pl)

distances :: (Int,Int) -> PointList -> [(Int,Pen)]
distances pt = map (oneDist pt)

oneDist :: (Int,Int) -> (Int,Int,Int,Int,Pen) -> (Int,Pen)
oneDist (xp,yp) (x,y,_,_,col) = ((dx*dx + dy*dy),col)
--if the distance to a point is greater than half the resolution, measures the distance by wrapping around instead
  where dx | abx > hDims = dims - abx
           | otherwise = abx
        dy | aby > hDims = dims - aby
           | otherwise = aby
        abx = abs(xp-x)
        aby = abs(yp-y)
        hDims = dims `div` 2

--uses recursion to find the colour value of the closest point
minDist :: [(Int,Pen)] -> Pen
minDist ((d,col):[]) = shaded d col
minDist (((d,col):(d1,col1):ps)) | abs(d1 - d) < 1 = minDist (((maximum[d,d1]),(avCol col col1)):ps)
                                 | d < d1 = minDist ((d,col):ps)
                                 | otherwise = minDist ((d1,col1):ps)

--antialising on the borders
avCol :: Pen -> Pen -> Pen
avCol (Colour r g b) (Colour r1 g1 b1) = (Colour ((r+r1)/2) ((g+g1)/2) ((b+b1)/2))

--colour gets darker the further from the point you are
shaded :: Int -> Pen -> Pen
shaded d (Colour r g b) | d <= 0   = white
                        | d <= 1   = black
                        | otherwise = Colour (r-df) (g-df) (b-df)
  where df = (sqrt(fromIntegral d)) * fadeOff